from enum import Enum
from typing import Any, List


class AutoValueEnum(Enum):
    def _generate_next_value_(  # type: ignore
        name: str, start: int, count: int, last_values: List[Any]
    ) -> Any:
        return name



from typing import Optional, Sequence
from urllib.parse import urlparse

from pydantic import model_validator

from stac_pydantic.api.item import Item
from stac_pydantic.api.links import Links
from stac_pydantic.item_collection import ItemCollection as BaseItemCollection
from stac_pydantic.links import Relations


class ItemCollection(BaseItemCollection):
    """
    https://github.com/radiantearth/stac-api-spec/blob/v1.0.0/fragments/itemcollection/README.md
    https://github.com/radiantearth/stac-api-spec/blob/v1.0.0/item-search/README.md#link-relations
    """

    features: Sequence[Item]
    links: Optional[Links] = None
    numberMatched: Optional[int] = None
    numberReturned: Optional[int] = None

    @model_validator(mode="after")
    def required_links(self) -> "ItemCollection":
        if self.links:
            links_rel = [link.rel for link in self.links.root]
            if Relations.self in links_rel:
                self_href = [link for link in self.links.root if link.rel == "self"][0]
                item_collection_type = urlparse(self_href.href).path.split("/")[-1]

                if item_collection_type == "items":
                    required_links = [
                        Relations.root,
                        Relations.self,
                        Relations.collection,
                    ]
                    for link in required_links:
                        assert (
                            link in links_rel
                        ), f"STAC API FEATURES conform Items pages must include a `{link}` link."

                if item_collection_type == "search":
                    required_links = [Relations.root]
                    for link in required_links:
                        assert (
                            link in links_rel
                        ), f"STAC API FEATURES conform Items pages must include a `{link}` link."

        return self



from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import Field

from stac_pydantic.links import Link
from stac_pydantic.links import Links as BaseLinks
from stac_pydantic.links import Relations
from stac_pydantic.shared import MimeTypes


class _MethodLink(Link):
    """
    https://github.com/radiantearth/stac-api-spec/blob/v1.0.0/item-search/README.md#pagination
    """

    method: Literal["GET", "POST"] = "GET"
    headers: Optional[Dict[str, Any]] = None
    body: Optional[Dict[Any, Any]] = None
    merge: bool = False


class PaginationLink(_MethodLink):
    """
    https://github.com/radiantearth/stac-api-spec/blob/v1.0.0/item-search/README.md#pagination
    """

    rel: Literal[Relations.next, Relations.prev]


class SearchLink(_MethodLink):
    """
    https://github.com/radiantearth/stac-api-spec/tree/v1.0.0/item-search#link-relations
    """

    href: str = Field(..., pattern=r".*\/search(\?.*)?\b")
    type: Literal[MimeTypes.geojson] = MimeTypes.geojson
    rel: Literal[Relations.self, Relations.search, Relations.next, Relations.prev]


class ItemsLink(_MethodLink):
    href: str = Field(..., pattern=r".*\/items(\?.*)?\b")
    type: Literal[MimeTypes.geojson] = MimeTypes.geojson
    rel: Literal[Relations.self, Relations.items]


class Links(BaseLinks):
    root: List[Union[SearchLink, PaginationLink, ItemsLink, Link]]



from stac_pydantic.scripts.cli import app


def test_valid_stac_item(cli_runner):
    result = cli_runner.invoke(
        app,
        [
            "validate-item",
            "https://raw.githubusercontent.com/radiantearth/stac-spec/v1.0.0/examples/extended-item.json",
        ],
    )

    if result.exception:
        raise result.exception

    assert result.exit_code == 0



from functools import lru_cache
from typing import Any, Dict, Union

from stac_pydantic.catalog import Catalog
from stac_pydantic.collection import Collection
from stac_pydantic.item import Item

try:
    import requests
except ImportError:  # pragma: nocover
    requests = None  # type: ignore

try:
    import jsonschema
except ImportError:  # pragma: nocover
    jsonschema = None  # type: ignore


@lru_cache(maxsize=128)
def _fetch_and_cache_schema(url: str) -> dict:
    """Fetch the remote JSON schema, if not already cached."""
    req = requests.get(url)
    return req.json()


def validate_extensions(
    stac_obj: Union[Item, Collection, Catalog, Dict[str, Any]],
    reraise_exception: bool = False,
) -> bool:
    """
    Fetch the remote JSON schema, if not already cached, and validate the STAC
    object against that schema.
    """
    assert requests is not None, "requests must be installed to validate extensions"
    assert jsonschema is not None, "jsonschema must be installed to validate extensions"

    if isinstance(stac_obj, dict):
        stac_dict = stac_obj
    else:
        stac_dict = stac_obj.model_dump(mode="json")

    try:
        if stac_dict["stac_extensions"]:
            for ext in stac_dict["stac_extensions"]:
                schema = _fetch_and_cache_schema(ext)
                jsonschema.validate(instance=stac_dict, schema=schema)
    except Exception:
        if reraise_exception:
            raise
        return False
    return True



import click
import requests
from pydantic import ValidationError

from stac_pydantic.extensions import validate_extensions
from stac_pydantic.item import Item

try:
    import requests
except ImportError:  # pragma: nocover
    requests = None  # type: ignore

try:
    import jsonschema
except ImportError:  # pragma: nocover
    jsonschema = None  # type: ignore


@click.group(short_help="Validate STAC")
def app() -> None:
    """stac-pydantic cli group"""
    pass


@app.command(short_help="Validate STAC Item")
@click.argument("infile")
def validate_item(infile: str) -> None:
    """Validate stac item"""
    assert requests is not None, "requests must be installed to validate items"
    assert jsonschema is not None, "jsonschema must be installed to validate items"

    r = requests.get(infile)
    r.raise_for_status()

    stac_item = r.json()
    try:
        item = Item.model_validate(stac_item)
        validate_extensions(item, reraise_exception=True)
    except ValidationError as e:
        click.echo(str(e))
        return

    click.echo(f"{infile} is valid")
    return



from datetime import datetime as dt
from typing import Any, Dict, List, Optional, Tuple, Union, cast

from geojson_pydantic.geometries import (
    GeometryCollection,
    LineString,
    MultiLineString,
    MultiPoint,
    MultiPolygon,
    Point,
    Polygon,
)
from pydantic import BaseModel, Field, TypeAdapter, field_validator, model_validator

from stac_pydantic.api.extensions.fields import FieldsExtension
from stac_pydantic.api.extensions.query import Operator
from stac_pydantic.api.extensions.sort import SortExtension
from stac_pydantic.shared import BBox, UtcDatetime

Intersection = Union[
    Point,
    MultiPoint,
    LineString,
    MultiLineString,
    Polygon,
    MultiPolygon,
    GeometryCollection,
]

SearchDatetime = TypeAdapter(Optional[UtcDatetime])


class Search(BaseModel):
    """
    The base class for STAC API searches.

    https://github.com/radiantearth/stac-api-spec/blob/v1.0.0/item-search/README.md#query-parameter-table
    """

    collections: Optional[List[str]] = None
    ids: Optional[List[str]] = None
    bbox: Optional[BBox] = None
    intersects: Optional[Intersection] = None
    datetime: Optional[str] = None
    limit: Optional[int] = 10

    # Private properties to store the parsed datetime values. Not part of the model schema.
    _start_date: Optional[dt] = None
    _end_date: Optional[dt] = None

    # Properties to return the private values
    @property
    def start_date(self) -> Optional[dt]:
        return self._start_date

    @property
    def end_date(self) -> Optional[dt]:
        return self._end_date

    # Check https://docs.pydantic.dev/dev-v2/migration/#changes-to-validators for more information.
    @model_validator(mode="before")
    def validate_spatial(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get("intersects") and values.get("bbox") is not None:
            raise ValueError("intersects and bbox parameters are mutually exclusive")
        return values

    @field_validator("bbox")
    @classmethod
    def validate_bbox(cls, v: BBox) -> BBox:
        if v:
            # Validate order
            if len(v) == 4:
                xmin, ymin, xmax, ymax = cast(Tuple[int, int, int, int], v)
            else:
                xmin, ymin, min_elev, xmax, ymax, max_elev = cast(
                    Tuple[int, int, int, int, int, int], v
                )
                if max_elev < min_elev:
                    raise ValueError(
                        "Maximum elevation must greater than minimum elevation"
                    )

            if xmax < xmin:
                raise ValueError(
                    "Maximum longitude must be greater than minimum longitude"
                )

            if ymax < ymin:
                raise ValueError(
                    "Maximum longitude must be greater than minimum longitude"
                )

            # Validate against WGS84
            if xmin < -180 or ymin < -90 or xmax > 180 or ymax > 90:
                raise ValueError("Bounding box must be within (-180, -90, 180, 90)")

        return v

    @field_validator("datetime")
    @classmethod
    def validate_datetime(cls, value: str) -> str:
        # Split on "/" and replace no value or ".." with None
        values = [v if v and v != ".." else None for v in value.split("/")]

        # If there are more than 2 dates, it's invalid
        if len(values) > 2:
            raise ValueError(
                "Invalid datetime range. Too many values. Must match format: {begin_date}/{end_date}"
            )

        # If there is only one date, duplicate to use for both start and end dates
        if len(values) == 1:
            values = [values[0], values[0]]

        # Cast because pylance gets confused by the type adapter and annotated type
        dates = cast(
            List[Optional[dt]],
            [
                # Use the type adapter to validate the datetime strings, strict is necessary
                # due to pydantic issues #8736 and #8762
                SearchDatetime.validate_strings(v, strict=True) if v else None
                for v in values
            ],
        )

        # If there is a start and end date, check that the start date is before the end date
        if dates[0] and dates[1] and dates[0] > dates[1]:
            raise ValueError(
                "Invalid datetime range. Begin date after end date. "
                "Must match format: {begin_date}/{end_date}"
            )

        # Store the parsed dates
        cls._start_date = dates[0]
        cls._end_date = dates[1]
        # Return the original string value
        return value

    @property
    def spatial_filter(self) -> Optional[Intersection]:
        """Return a geojson-pydantic object representing the spatial filter for the search request.

        Check for both because the ``bbox`` and ``intersects`` parameters are mutually exclusive.
        """
        if self.bbox:
            return Polygon.from_bounds(*self.bbox)
        if self.intersects:
            return self.intersects
        else:
            return None


class ExtendedSearch(Search):
    """
    STAC API search with extensions enabled.
    """

    field: Optional[FieldsExtension] = Field(None, alias="fields")
    query: Optional[Dict[str, Dict[Operator, Any]]] = None
    sortby: Optional[List[SortExtension]] = None



from typing import List, Literal, Optional

from pydantic import AnyUrl, ConfigDict, Field

from stac_pydantic.links import Links
from stac_pydantic.shared import SEMVER_REGEX, StacBaseModel
from stac_pydantic.version import STAC_VERSION


class _Catalog(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/catalog-spec/catalog-spec.md
    """

    id: str = Field(..., alias="id", min_length=1)
    description: str = Field(..., alias="description", min_length=1)
    stac_version: str = Field(STAC_VERSION, pattern=SEMVER_REGEX)
    links: Links
    stac_extensions: Optional[List[AnyUrl]] = []
    title: Optional[str] = None
    type: str
    model_config = ConfigDict(use_enum_values=True, extra="allow")


class Catalog(_Catalog):
    type: Literal["Catalog"]



from datetime import timezone
from enum import Enum, auto
from typing import Any, Dict, List, Optional, Tuple, Union
from warnings import warn

from pydantic import (
    AfterValidator,
    AwareDatetime,
    BaseModel,
    ConfigDict,
    Field,
    model_validator,
)
from typing_extensions import Annotated, Self

from stac_pydantic.utils import AutoValueEnum

NumType = Union[float, int]
BBox = Union[
    Tuple[NumType, NumType, NumType, NumType],  # 2D bbox
    Tuple[NumType, NumType, NumType, NumType, NumType, NumType],  # 3D bbox
]

SEMVER_REGEX = r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$"

# Allows for some additional flexibility in the input datetime format. As long as
# the input value has timezone information, it will be converted to UTC timezone.
UtcDatetime = Annotated[
    # Input value must be in a format which has timezone information
    AwareDatetime,
    # Convert the input value to UTC timezone
    AfterValidator(lambda d: d.astimezone(timezone.utc)),
]


class MimeTypes(str, Enum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#media-types
    """

    # Raster
    geotiff = "image/tiff; application=geotiff"
    cog = "image/tiff; application=geotiff; profile=cloud-optimized"
    jp2 = "image/jp2"
    png = "image/png"
    jpeg = "image/jpeg"
    # Vector
    geojson = "application/geo+json"
    geojsonseq = "application/geo+json-seq"
    geopackage = "application/geopackage+sqlite3"
    kml = "application/vnd.google-earth.kml+xml"
    kmz = "application/vnd.google-earth.kmz"
    pbf = "application/x-protobuf"
    mvt = "application/vnd.mapbox-vector-tile"
    # Others
    hdf = "application/x-hdf"
    hdf5 = "application/x-hdf5"
    xml = "application/xml"
    json = "application/json"
    ndjson = "application/ndjson"
    html = "text/html"
    text = "text/plain"
    openapi = "application/vnd.oai.openapi+json;version=3.0"
    openapi_yaml = "application/vnd.oai.openapi;version=3.0"
    jsonschema = "application/schema+json"
    pdf = "application/pdf"
    csv = "text/csv"
    parquet = "application/vnd.apache.parquet"


class AssetRoles(str, AutoValueEnum):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/extensions/asset/README.md
    """

    thumbnail = auto()
    overview = auto()
    data = auto()
    metadata = auto()


class ProviderRoles(str, AutoValueEnum):
    licensor = auto()
    producer = auto()
    processor = auto()
    host = auto()


class StacBaseModel(BaseModel):
    def to_dict(
        self, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> Dict[str, Any]:
        warn(
            "`to_dict` method is deprecated. Use `model_dump` instead",
            DeprecationWarning,
        )
        return self.model_dump(by_alias=by_alias, exclude_unset=exclude_unset, **kwargs)

    def to_json(
        self, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> str:
        warn(
            "`to_json` method is deprecated. Use `model_dump_json` instead",
            DeprecationWarning,
        )
        return self.model_dump_json(
            by_alias=by_alias, exclude_unset=exclude_unset, **kwargs
        )

    def model_dump(
        self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> Dict[str, Any]:
        return super().model_dump(
            by_alias=by_alias, exclude_unset=exclude_unset, **kwargs
        )

    def model_dump_json(
        self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: Any
    ) -> str:
        return super().model_dump_json(
            by_alias=by_alias, exclude_unset=exclude_unset, **kwargs
        )


class Provider(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/collection-spec/collection-spec.md#provider-object
    """

    name: str = Field(..., min_length=1)
    description: Optional[str] = None
    roles: Optional[List[str]] = None
    url: Optional[str] = None


class StacCommonMetadata(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/common-metadata.md
    """

    # Basic
    title: Optional[str] = None
    description: Optional[str] = None
    # Date and Time
    datetime: Optional[UtcDatetime] = Field(...)
    created: Optional[UtcDatetime] = None
    updated: Optional[UtcDatetime] = None
    # Date and Time Range
    start_datetime: Optional[UtcDatetime] = None
    end_datetime: Optional[UtcDatetime] = None
    # Licensing
    license: Optional[str] = None
    # Provider
    providers: Optional[List[Provider]] = None
    # Instrument
    platform: Optional[str] = None
    instruments: Optional[List[str]] = None
    constellation: Optional[str] = None
    mission: Optional[str] = None
    gsd: Optional[float] = Field(None, gt=0)

    @model_validator(mode="after")
    def validate_datetime_or_start_end(self) -> Self:
        # When datetime is null, start_datetime and end_datetime must be specified
        if not self.datetime and (not self.start_datetime or not self.end_datetime):
            raise ValueError(
                "start_datetime and end_datetime must be specified when datetime is null"
            )

        return self

    @model_validator(mode="after")
    def validate_start_end(self) -> Self:
        # Using one of start_datetime or end_datetime requires the use of the other
        if (self.start_datetime and not self.end_datetime) or (
            not self.start_datetime and self.end_datetime
        ):
            raise ValueError(
                "use of start_datetime or end_datetime requires the use of the other"
            )
        return self


class Asset(StacBaseModel):
    """
    https://github.com/radiantearth/stac-spec/blob/v1.0.0/item-spec/item-spec.md#asset-object
    """

    href: str = Field(..., min_length=1)
    type: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    roles: Optional[List[str]] = None

    model_config = ConfigDict(
        populate_by_name=True, use_enum_values=True, extra="allow"
    )
